package ZipTie::Adapters::CheckPoint::OPSEC::Parsers;

use 5.010_000;
use strict;
use warnings;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use Net::CIDR::Lite;
use Net::CIDR::Set;
use ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser;
use ZipTie::Adapters::ConfigObjects;
use ZipTie::Adapters::Utils qw(get_crep get_port_number mask_to_bits trim get_interface_type get_any_address_encoding get_protocol_name);
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(checkNAT get_any parse_rules parse_object_groups parse_interfaces parse_rules_file generate_filter_list object_to_hash);

my $LOGGER = ZipTie::Logger::get_logger();

sub parse_rules {	
	my ($in, $out) = @_;

	$out->open_element("filterLists");
	
	# check NAT existence
	checkNAT($in,$out); 
	
	if ($in->{isNAT}){
		# print out packetFiltering
		# this happens before translating and routing regardless of 
		# 'translate destination on client side' (nat_dst_client_side) switch value from properties.xml
		$out->open_element('filterListsSection');	
			_packetFiltering($in,$out);
		$out->close_element('filterListsSection');

		# print out address-translation
		_addressTranslation ($in, $out);
	} else {
		_packetFiltering($in,$out);
	}

	$out->close_element("filterLists");
	
}

sub _addressTranslation {
	
	my ($in, $out) = @_;
	
	#set SecurePlatform environment	
	$in->{isCPSMS} = 0;
	
	# set manual NAT as undefined by default
	$in->{manualNAT}->{available} = 0;
	
	# get an addressTranslationParser object
	my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
	# store global NAT settings under $in->{NAT}->{settings}
	$addressTranslationParser->getNATProperties();
	
	#set processingPhase
	$in->{NAT}->{settings}->{processingPhase} = "address-translation";
	my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
	
	# check where the automatic NAT is applied, at the server side or at the client side
	if ( defined($in->{NAT}->{settings}->{nat_dst_client_side}) and  $in->{NAT}->{settings}->{nat_dst_client_side} =~ /(t)|(T)rue/)  {
		$out->open_element('filterListsSection');
			# if translation is executed at the client side 
			my $routingNATFilterList = $addressTranslationParser->routingNAT('automatic');
			$out->print_string($routingNATFilterList);	
			# add the phase discriminator
			$in->{NAT}->{settings}->{processingPhase} = 'routing';
			$in->{NAT}->{settings}->{processingPhasePosition} = 'before';
			my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
			$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');		
	} else {
		$out->open_element('filterListsSection');
			# else translation is executed at the server side
			my $routingNATFilterList = $addressTranslationParser->routingNAT('automatic');
			$out->print_string($routingNATFilterList);
			# add the phase discriminator
			$in->{NAT}->{settings}->{processingPhase} = 'routing';
			$in->{NAT}->{settings}->{processingPhasePosition} = 'after';
			my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
			$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');
	}
	
	#check if manual NAT is defined
	$addressTranslationParser->checkManualNAT();
	if (defined ($in->{manualNAT}->{available}) and $in->{manualNAT}->{available} == 1){
		#check where the manual NAT is applied, at the server side or at the client side
		if ( defined($in->{NAT}->{settings}->{nat_dst_client_side_manual}) and $in->{NAT}->{settings}->{nat_dst_client_side_manual} =~ /(t)|(T)rue/){
			$out->open_element('filterListsSection');
				# if translation is executed at the client side
				my $routingNATFilterList = $addressTranslationParser->routingNAT('manual');
				$out->print_string($routingNATFilterList);	
				# add the phase discriminator
				$in->{NAT}->{settings}->{processingPhase} = 'routing';
				$in->{NAT}->{settings}->{processingPhasePosition} = 'before';
				my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
				$out->print_string($phaseDiscriminatorPhase);		
			$out->close_element('filterListsSection');	
		} else {
			$out->open_element('filterListsSection');
				# else translation is executed at the server side
				my $routingNATFilterList = $addressTranslationParser->routingNAT('manual');
				$out->print_string($routingNATFilterList);
				# add the phase discriminator
				$in->{NAT}->{settings}->{processingPhase} = 'routing';
				$in->{NAT}->{settings}->{processingPhasePosition} = 'after';
				my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
				$out->print_string($phaseDiscriminatorPhase);	
			$out->close_element('filterListsSection');
		}
	} # check manual NAT existence
			
	
}

sub _packetFiltering {
	my ( $in, $out ) = @_;

	my $filterList = {
		mode => 'stateful',
	};

	# if NAT configuration is enabled then set processingPhase
	if ($in->{isNAT}) {
		$in->{NAT}->{settings}->{processingPhase} = "packet-filtering";
	}

	# Call on opsec rules first just because it means we'll get interface association to handle 
	# better
	my $opsecFeSets = parse_rules_file({in => $in, 
		rules_file => $in->{rulesFile}, 
		filterList => $filterList});

	my $rulesTotal = _count_rule_set($opsecFeSets);

	$LOGGER->debug("CheckPoint::OPSEC::parse_rules: $rulesTotal user rules found and parsed");

	# note bene: we could be carrying around the $appliesTo as an undef element.	
	my $appliesTo = $filterList->{appliesTo} // undef;

	$in->{enabled_services} = _get_service_enablement($in);

	my $impliedFeSets = parse_rules_file({in => $in, 
				rules_file => $in->{impliedFile}, 
				filterList => $filterList, 
				rules_only => 1});

	$rulesTotal = _count_rule_set($impliedFeSets);
	$LOGGER->debug("CheckPoint::OPSEC::parse_rules: $rulesTotal implied found and parsed");

	delete $in->{enabled_services};

	# Now, let's generate the lists.  We do this in the following order:
	#   * implied rules, "first" clauses
	#   * opsec rules (which should all be unprioritized)
	#   * unprioritized implied rules
	#   * 'before last' implied rules
	#   *  'last' implied rules
	#
	# We make sure to have a default disposition of "none" for all lists right up to the end

	generate_filter_list( { out => $out, filterEntries => $impliedFeSets->{'first'}, 
			filterName => "Implied-first", filterDesc => "Global rules - 'first' entries", 
			applicableInterfaces => $appliesTo, defaultAction => "none", in => $in});
	generate_filter_list({ out => $out,filterEntries =>  $opsecFeSets->{'unordered'}, 
			filterName => "OPSEC", filterDesc => "OPSEC user rules", 
			applicableInterfaces => $appliesTo, defaultAction => "none", in => $in});
	generate_filter_list({out => $out, filterEntries =>  $impliedFeSets->{'unordered'}, 
			filterName => "Implied", filterDesc => "Global unordered rules", 
			applicableInterfaces => $appliesTo, defaultAction => "none", in => $in});
	generate_filter_list({out => $out,filterEntries =>  $impliedFeSets->{'before last'}, 
			filterName => "Implied-before-last", 
			filterDesc => "Global rules to apply before default last rules", 
			applicableInterfaces => $appliesTo, defaultAction => "none", in => $in});
	generate_filter_list({out => $out, filterEntries =>  $impliedFeSets->{'last'}, 
			filterName => "Implied-last", filterDesc =>  "Global default last rules", 
			applicableInterfaces => $appliesTo, 
			defaultAction => "drop", forceGeneration => 1, in => $in});

	# I don't think we need this any more since 'defaultAction' on the final filter will take care of it						  
	#    my $defaultFilter = _add_default_filter($filterList);
	#    push(@{$filterList->{filterEntry}}, $defaultFilter); 

	#    $out->close_element('filterLists');

	# can clean up accumulated state built up for rule checking
	# keep networkGroups and portServiceGroups
	# it is used as reference within address translantion part 
	# delete $in->{networkGroups} if defined ($in->{networkGroups});
	# delete $in->{portServiceGroups} if defined ($in->{portServiceGroups});
	delete $in->{ipProtocolGroups} if defined ($in->{ipProtocolGroups});
	return;
}

sub generate_filter_list
{
	my ($args) = @_;
	#
	#  Named arg list:
	#    {out} - XMLPrinter instance being streamed to
	#    {filterEntries} - entries to put into generated filter list.  Can be empty list.
	#    {filterName} - name to apply to filter
	#    {filterDesc} - description of filter
	#    {applicableInterfaces} - interfaces to which to apply this filter list
	#    {defaultAction} - default for filter list if nothing matches in eval
	#    {forceGeneration} - optional (default false) - if true then force generation of empty 
	#              filter list even if $args->{filterList} is an empty list
	#    {in} - optional (default undefined) - if true then input configuration is passed
	
	$args->{forceGeneration} = (defined ($args->{forceGeneration})) ? $args->{forceGeneration} : 0;
	$args->{in} = (defined ($args->{in})) ? $args->{in} : undef;
	
	my $processIndex = 1; # first assigned processOrder element is 1.
	my $filterList;
	
	if ((not defined $args->{filterEntries}) or (0 == scalar(@{$args->{filterEntries}})))
	{
		# in case an empty hash bucket or entry list gets passed, only move forward if we're
		# forcin' it
		return unless $args->{forceGeneration};
	}
		
	# owing to out-of-order fields alphabetically (grrr!!!) we have to generate the fields directly
	$args->{out}->open_element('filterList');
	$args->{out}->print_element('defaultAction', $args->{defaultAction});
	$args->{out}->print_element('description', $args->{filterDesc}) if  $args->{filterDesc};
	$args->{out}->print_element('appliesTo', $args->{applicableInterfaces}) if $args->{applicableInterfaces};
	
	for my $fe (@{$args->{filterEntries}})
	{
 	   	$fe->{processOrder} = $processIndex++;
 	   	$args->{out}->print_element('filterEntry', $fe);
	}

	$args->{out}->print_element('mode', 'stateful');
	$args->{out}->print_element('name', $args->{filterName});
	# if NAT is enabled then add the packetProcessingPhase element for packet-filtering phase
	if ($args->{in}->{isNAT}) {
		# get an addressTranslationParser object
		my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($args->{in});
		my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
		$args->{out}->print_string($packetProcessingPhase);
	} 
	$args->{out}->close_element('filterList');		
}

sub _add_default_filter
{
    my ($out) = @_;

    my $drop = {
        'name' => 'default_action',
        'log' => 'false',
        'primaryAction' => 'deny',
    };

    push (@{$drop->{'sourceIpAddr'}}, get_any());
    push (@{$drop->{'destinationIpAddr'}}, get_any());

    return $drop;
}

sub parse_rules_file
{
	my ($args) = @_;

	# Args are:
	#   {in} - configuration
	#   {rules_file} - file to check for rules
	#   {filterList} - filter list header
	#   {rules_only} - optional (false if not specified) : just check for rules section, not others
	#   {impliedFeSets} - optional (false if not specified) : used to generate recimpliedFeSets file 
	#
	#  This returns a hash of filter entries, prioritized by any enabling_prop ordering we
	#  can dig up (first, last, etc.)  For rules which don't have such an ordering (indeed, 
	#  OPSEC filters don't at all) put into bucket 'unordered'.
	#
    #  With that in mind, we will create a position ordering of the filters we encounter...
    my $orderedFilters = { "first" => [], "unordered" => [],  "before last" => [], "last" => [] };

	$args->{rules_only} = defined($args->{rules_only}) ? $args->{rules_only} : 0;
	# set full path's name for impliedFeSets 
	$args->{impliedFeSets} = defined ($args->{impliedFeSets}) ? $args->{impliedFeSets} : 0;

	open( RULES, $args->{rules_file} ) or die("Unable to open file '$args->{rules_file}': $!");
	my $rulesWhitespace;          # leading whitepsace on the rules definition
	my $currentRuleWhiteSpace;    # leading whitespace on each rule definition
	my $currentRule;
    my $type;
    my $processIndex = 0;
    my $rPos;
    my $fe;

    if ($args->{rules_only})
    {
        $rulesWhitespace = "";
        $type = 'rules';
    }

	while ( my $line = <RULES> )
	{
		if( $line =~ /^\s*#.*$/ )
		{
			next;
		}
		if (defined($rulesWhitespace))
		{
			if ( $line =~ /^$rulesWhitespace\)/ )
			{
				$rulesWhitespace = undef;
			    $type = undef;
			}
			elsif ( ( $line =~ /(^\s+):\s*\(/ ) && ( !$currentRuleWhiteSpace ) )
			{
				$currentRuleWhiteSpace = $1;
				$currentRule .= $line;
			}
			elsif ( $currentRuleWhiteSpace && ( $line =~ /^($currentRuleWhiteSpace|\t)\)/ ) )
			{
			    $currentRule .= $line;
			    if ($type eq 'rules')
			    {
    				($fe, $rPos) = _process_rule( $currentRule, $args->{in} );
     				if ($fe)
    				{
    					# we have to find which of several ordered lists to put this on
    					if (not defined $rPos)
    					{
    						$rPos = "unordered";
    					}
    					
    					push (@{$orderedFilters->{"$rPos"}}, $fe);
    					if (defined $args->{in}->{propertylist}->{name}){
    					  	# push FilterEntry into propertylist
    						my $propname = $args->{in}->{propertylist}->{name};
    						delete $args->{in}->{propertylist}->{name};
    						push (@{$args->{in}->{propertylist}->{$propname}},$fe);
    					}
    				}
			    }
			    elsif ($type eq 'if_info')
			    {
					my $processedAlready = 0;	# eliminate duplication of appliesTo elements
					my @appliesTo = @{ _process_if_info($currentRule, $args->{in} ) };

			        foreach my $appliesToEntry (@appliesTo)
			        {
			        	if (exists($args->{filterList}->{appliesTo}))
			        	{
			        		foreach my $source (@{$args->{filterList}->{appliesTo}})
			        		{
			        			if ($appliesToEntry->{source} eq $source->{source})
			        			{
			        				$processedAlready = 1;	
			        			}
			        		}
			        	}
			        	if ($processedAlready == 0)
			        	{
			        		push(@{$args->{filterList}->{appliesTo}}, $appliesToEntry);
			        	}
			        }
			    }
				$currentRule           = undef;
				$currentRuleWhiteSpace = undef;
			}
			elsif ($line =~ /^\(/)
			{
			    next;
			}
			elsif ($line =~ /^#/)
			{
			    next;
			}
			else
			{
				$currentRule .= $line;
			}
		}
		elsif ( $line =~ /(^\s+):rules\s+/ )
		{
			$rulesWhitespace = $1;
			$type = 'rules';
		}
		elsif ( $line =~ /(^\s+):if_info\s+/ )
		{
		    $rulesWhitespace = $1;
		    $type = 'if_info';
		}
	}

	close(RULES);
	# save the newly recorded impliedFeSets
	if ($args->{impliedFeSets}){
		my $impliedFeSets = $args->{impliedFeSets};
    	open(FILE, ">$impliedFeSets") or $LOGGER->fatal_error_code("ERROR Couldn't open file $impliedFeSets..."); 
		print FILE Dumper ($args->{in}->{propertylist});
		close FILE;
	}    					

	
	return $orderedFilters;
}

#
# Gets a hash of enablement properties to array of services
sub _get_service_mappings
{
    my ($in) = @_;

    open(SETUP, $in->{setupFile}) or die;

    my $result;
    my $parent_ws;
    my $current_ws;
    my $current;
    my @alias;

    while (my $line = <SETUP>)
    {
        if ($parent_ws)
        {
            if ($line =~ /^$parent_ws\)/)
            {
                $parent_ws = undef;
                next;
            }
            elsif ($current_ws)
            {
                if ($line =~ /^$current_ws\)/)
                {
                    $current_ws = undef;
                    next;
                }
                # parse some props
                if ($line =~ /^\s+:(.+?)\s+\((.+?)\)/)
                {
                    my $prop = $1;
                    my $value = $2;

                    $value =~ s/^"//;
                    $value =~ s/"$//;

                    $current->{$prop} = $value;
                }
            }
            elsif ($line =~ /(^\s+):(.+?)\s+\(/)
            {
                $current_ws = $1;
                $current = {};
                push(@{$result->{$2}}, $current);
            }
        }
        elsif ($line =~ /(^\s+):(prop2pseudo)/)
        {
            $parent_ws = $1;
        }
    }

    close(SETUP);

    return $result;
}

# This is a catalog of observed deviations or bugs from the canonical finding of any given
# enabling service having a companion property in objects.C, such that in implied_rules.C 
#          :foo_rule ( enabling_prop(service) )
#
#        having a property in objects.C
#			:service (true)
#
#       AND having possibly a companion property
#           :service_p (first)   # or 'last' or 'before last'
#
our %AlternateServicePositionalPropertyMappings = (
		"fw1_enable_p" => "fw1enable",
		"r_access_enable_p" => "raccessenable",  
);


#
# Returns a hash where the keys are the names of "implied rules" "that are enabled.
sub _get_service_enablement
{
    my ($in) = @_;

    open(OBJECTS, $in->{objectsFile}) or die;

    my $result;

    my $parent_ws;
    my $type;
    my $servobj;
    my @falseEnabs;
    my @trueEnabs;

    while (my $line = <OBJECTS>)
    {
    	# Handle R70 comments convention
 		if( $line =~ /^\s*#.*$/ )
		{
			next;
		}
 
        if ($parent_ws)
        {
            if ($line =~ /^$parent_ws\)/)
            {
            	# We've unindented, and hence the 'props' section has closed
                $parent_ws = undef;
                next;
            }
            elsif ($type eq 'props')
            {
                if ($line =~ /^\s+:(\S+)\s+\((true|false)\)/)
                {
                    my $service = $1;
                    my $enabled = $2;

		    if ($enabled eq 'true')
		    {
			push (@trueEnabs, $service);
		    }
		    else
		    {
			push (@falseEnabs, $service);
		    }
                    next unless ($enabled eq 'true');

                    $result->{$service}->{enabled} = 1;
                }
                # look for a "foo_p" that has a positional designator, stripping the possible
                # quotes around that designator
                elsif (($line =~ /^\s+:(\S+?)_p\s+\(\"?(.*?)\"?\)/) and (defined $result->{$1}))
                {
                	# see if we have the "position" designator for a prior service
                	if (defined $result->{$1})
                	{
                		$result->{$1}->{position} = $2;
                	}
                }
                # else, check "alternative" mappings
                elsif (($line =~ /^\s+:(\S+)\s+\(\"?(.*?)\"?\)/) and (defined ($AlternateServicePositionalPropertyMappings{$1})))
                {
                	if (not defined ($result->{$AlternateServicePositionalPropertyMappings{$1}}->{position}))
                	{
                		$result->{$AlternateServicePositionalPropertyMappings{$1}}->{position} = $2;
                	}
                }
            }
        }
        elsif ($line =~ /(^\s+):(props)\s+\((\S+)/)
        {
            $parent_ws = $1;
            $type = $2;
            $servobj = $3;
        }
    }

		   
    close(OBJECTS);

    return $result;
}

# Parse network groups from Standard.pf
sub _parse_object_groups_from_standard_file {
	my ($in) = @_;
	my $ip_regex = get_crep('cipm');
	my %network_groups = ();

	if( open my $STANDARD, '<', $in->{StandardPF} ) {
		while( my $line = <$STANDARD> ) {

			# IPv6 lists end with '_list6', so this won't capture any
			if( $line =~ /^(\S+)_list = {(.+)};/ ) {
				my ($id, $addresses) = ($1, $2);
				my $network_group = { id => $id };

				while( $addresses =~ /<(\S+),\s*(\S+)>/g ) {
					my ($start, $end) = ($1, $2);

					if( $start =~ /$ip_regex/ ) {
						if( $end !~ /$ip_regex/ or $start eq $end ) {
							push @{ $network_group->{host} }, $start;
						} else {
							push @{ $network_group->{range} }, {startAddress => $start, endAddress => $end};
						}
					}
				}
				$network_groups{$id} = $network_group;
			}
		}
		close $STANDARD;
	} else {
		$LOGGER->debug( "Could not open $in->{StandardPF} to parse network objects: $!" );
	}

	return \%network_groups;
}

# Parse network, service and protocol groups from objects.C
sub _parse_object_groups_from_objects_file {
	my ($in) = @_;
	my %objects;

	if( open my $OBJECTS, '<', $in->{objectsFile} ) {
		my $parentWhitespace;        # leading whitepsace on the netobj definition
		my $currentObjectWhitespace; # leading whitespace on each object definition
		my $currentObject;
		my $type;

		while( my $line = <$OBJECTS> ) {

			# Handle comments
			next if $line =~ /^\s*#/;
			
			if( $line =~ /^(\s+):(netobj|servobj)\s*\(/ ) {
				$parentWhitespace = $1;
				$type             = $2;
			} elsif( $parentWhitespace ) {

				# End of the netobj or servobj block, reset parent flag
				if( $line =~ /^$parentWhitespace\)/ ) {
					$parentWhitespace = undef;

				# Start of a new object, grab whitespace and start appending object
				} elsif( !$currentObjectWhitespace and $line =~ /^(\s+):\s*\(/ ) {
					$currentObjectWhitespace = $1;
					$currentObject .= $line;

				# End of current object, parse and reset current object state
				} elsif( $currentObjectWhitespace and $line =~ /^$currentObjectWhitespace\)/ ) {
					my $current = object_to_hash($currentObject);
					my ($name)  = keys %{ $current };
					my $object  = $current->{$name};

					if( $object ) {
						if( $type eq 'netobj' ) {

							# Parse data into $in for later use, if this is a host/gw/gw cluster
							_parse_checkpoint_objects($in, $object, $name);

							my $network_group = _parse_network_group( $in, $object, $name );

							# Store group with exclusions later, the rest of the network groups need
							# parsed in order to resolve the exclusions.
							if( $network_group ) {
								if( $network_group->{base} ) {
									$objects{group_with_exclusion}->{$name} = $network_group;
								} else {
									$objects{network_groups}->{$name} = $network_group;
								}
							}

						} elsif( $type eq 'servobj' ) {
							my $protocol_group = _parse_protocol_group( $in, $object, $name );
							$objects{protocol_groups}->{$name} = $protocol_group if $protocol_group;

							my ($service_group, $src_service_group) = _parse_service_group( $in, $object, $name );
							$objects{service_groups}->{$name} = $service_group if $service_group;
							$objects{service_groups}->{$name .'__src'} = $src_service_group if $src_service_group;
						}
					}
					$currentObject           = undef;
					$currentObjectWhitespace = undef;
				} else {
					$currentObject .= $line;
				}
			}
		}
		close $OBJECTS;
	} else {
		$LOGGER->debug( "Could not open $in->{objectsFile} to parse object groups: $!" );
	}

	return (\%objects);
}

sub parse_object_groups {
	my ($in, $out) = @_;
	my %objects;

	$objects{network_groups} = _parse_object_groups_from_standard_file($in);

	my $objects_from_file = _parse_object_groups_from_objects_file($in);
	$objects_from_file->{network_groups} = { %{$objects{network_groups}}, %{ $objects_from_file->{network_groups} } };
	%objects = ( %objects, %{ $objects_from_file} );

	# Resolve group with exclusions
	for my $name ( keys %{ $objects{group_with_exclusion} } ) {
		my $network_group = { id => $name };
		my $group_with_exclusion = $objects{group_with_exclusion}->{$name};
		my $base = $group_with_exclusion->{base};
		my $base_set;
		if( $base =~ /any/i ) {
			$base_set = Net::CIDR::Set->new('0.0.0.0/0');
		} else {
			$base_set = _get_network_set($group_with_exclusion->{base}, $objects{network_groups});
		}

		if( $base_set ) {
			my $exclusion = _get_network_set($group_with_exclusion->{exclusion}, $objects{network_groups});
			if( $exclusion ) {
				my $diff = $base_set->diff($exclusion);
				my @diffs = $diff->as_range_array();
				if( @diffs ) {
					for my $diff( @diffs ) {
						if( $diff =~ /(.+)-(.+)/ ) {
							push @{ $network_group->{range} }, { startAddress => $1, endAddress => $2 };
						} else {
							my ($address, $mask) = split /\//, $diff;
							push @{ $network_group->{subnet} }, { address => $address, mask => $mask // 32 };
						}
					}
					$objects{network_groups}->{$name} = $network_group;
				}
			} else {
				$LOGGER->debug("WARNING: Dropping network group. Unknown 'group with exclusion' exclusion '$group_with_exclusion->{exclusion}' for '$name'");
			}
		} else {
			$LOGGER->debug("WARNING: Dropping network group. Unknown 'group with exclusion' base '$base' for '$name'");
		}
	}
	delete $objects{group_with_exclusion};

	# Create a network group for all the hosts with Check Point modules installed
	if( $in->{cp_products_hosts} ) {
		$objects{network_groups}->{cp_products} = {
			id => 'cp_products',
			host => $in->{cp_products_hosts}
		};
	}

	# Create implied 'my_ip' and 'my_ip_and_local_cluster' network groups
	if( $in->{my_gateway_metadata} ) {
		$objects{network_groups}->{my_ip} = {
			id => 'my_ip',
			host => $in->{my_gateway_metadata}->{my_ip}
		};

		$objects{network_groups}->{my_ip_and_local_cluster} = _build_my_ip_and_local_cluster($in);
	}

	# Add all the parsed network group names to a list, to check for missing references
	$in->{networkGroups} = $objects{network_groups};

	_print_object_groups($in, $out, \%objects);

	return;
}

# Handles printing of object groups that have been parsed into arrays
sub _print_object_groups {
	my ( $in, $out, $objects ) = @_;
	my %objects = %{ $objects };
	my @network_groups = keys %{ $objects{network_groups} };
	my @protocol_groups = keys %{ $objects{protocol_groups} };
	my @service_groups = keys %{ $objects{service_groups} };

	$out->open_element('objectGroups');

	# Print the network groups, unless they reference a group that does not exist
	# then drop that reference and only print the group if it still has valid refernces
	for my $name ( @network_groups ) {
		my $network_group = $objects{network_groups}->{$name};
		if( $network_group->{objectGroupReference} ) {
			my @object_references = @{ $network_group->{objectGroupReference} };
			delete $network_group->{objectGroupReference};

			for my $object_reference ( @object_references ) {
				if( $in->{networkGroups}->{ $object_reference } ) {
					push @{ $network_group->{objectGroupReference} }, $object_reference;
				} else {
					$LOGGER->debug("[WARNING] Network group reference not found '$object_reference'");
				}
			}

			if( $network_group->{objectGroupReference} ) {
				_print_network_group( $out, $network_group );
			} else {
				delete $in->{networkGroups}->{$network_group->{id}};
			}
		} else {
			_print_network_group( $out, $network_group );
		}
	}

	for my $name ( @protocol_groups ) {
		my $protocol_group = $objects{protocol_groups}->{$name};
		if( defined $protocol_group->{protocol} ) {
			$out->print_element( 'protocolGroup', $protocol_group );
		} else {
			push @service_groups, $protocol_group;
		}
	}

	for my $name ( @service_groups ) {
		my $service_group = $objects{service_groups}->{$name};
		$out->print_element( 'serviceGroup', $service_group );
	}

	$out->close_element('objectGroups');

	return;
}

#  _build_my_ip_and_local_cluster
#
#  This creates a network group, called "my_ip_and_local_cluster", out of available information we can
#  find on high availabity clustering in the configuration.  It relates the relationships between IP addresses
#  in a cluster where one of those represents a local secure gateway IP.
sub _build_my_ip_and_local_cluster {
	my ($in) = @_;
	my $network_group = { id => 'my_ip_and_local_cluster' };

	my %hosts = ( $in->{my_gateway_metadata}->{my_ip} => 1 );
	if( $in->{clusters} ) {
		my $name = $in->{my_gateway_metadata}->{name};

		# Collect all ip addresses of interfaces, in clusters who contain the local gateway
		while( my ($cluster_name, $cluster_data) = each %{ $in->{clusters} } ) {
			foreach my $member_name ( @{ $cluster_data->{members} } ) {
				if( $member_name eq $name ) {
					while( my ($interface_name,$interface_data) = each %{$cluster_data->{interfaces}} ) {
						$hosts{$interface_data->{ipaddr}} = 1;
					}
				}
			}
		}

	}

	my @hosts = keys %hosts;
	$network_group->{host} = \@hosts;

	return $network_group;
}

sub _process_if_info {
	my ($config, $in) = @_;
	my @appliesTo;

	my $obj = _rule_to_hash($config);
	foreach my $info (@$obj) {
		foreach my $child (@{$info->{children}}) {
			if (defined $child->{name}) {
				next;
			}
			push(@appliesTo, {source => $child->{simple}});
		}
	}

	return \@appliesTo;
}


sub _process_rule
{
	my ( $ruleConfig, $in ) = @_;
	my $ruleObj = _rule_to_hash($ruleConfig);
	$ruleObj = @$ruleObj[0];
	my $name = $ruleObj->{simple};
	my $rPos = undef;

	my $action;
	my $filterEntry = { name => $name, log => 'false', };

	foreach my $child ( @{ $ruleObj->{children} } )
	{
	    my $cname = $child->{name};
	    if( $cname eq 'AdminInfo' )
	    {
	    	foreach my $adminInfo ( @{ $child->{children} } )
	    	{
	    		if( $adminInfo->{name} eq 'chkpf_uid' )
	    		{
	    			$filterEntry->{ztSignature} = $adminInfo->{simple};
	    		}
	    	}
	    }
	    elsif ($cname eq 'track')
	    {
	        foreach my $track ( @{ $child->{children} } )
	        {
	            if (defined($track->{simple}) and $track->{simple} eq 'Log')
	            {
	                $filterEntry->{log} = 'true'
	            }
	        }
	    }
	    elsif ($cname eq 'enabling_prop')
	    {
	        my $c = @{ $child->{children} }[0];
	        my $simple = $c->{simple};

	        unless ($simple and exists $in->{enabled_services}->{$simple})
		{
		    # Rule not enabled.
		    return (undef, 0);
		}

		# keep for record the property name 
		$in->{propertylist}->{name} = $simple;
            
		if (defined $in->{enabled_services}->{$simple}->{position})
		{
		    $rPos = $in->{enabled_services}->{$simple}->{position};
		    push (@{$in->{service_enabs}->{"$simple"}}, $name);
		}
	    }
	    elsif ($cname eq 'action')
	    {
	        my $primary = @{ $child->{children} }[0];
            $action = _get_simple_by_name($primary->{children}, 'action');
            unless ($action)
            {
                $action = _get_simple_by_name($primary->{children}, 'type');
                unless ($action)
                {
                    $action = $primary->{simple};
                }
            }
	    }
	    elsif ($cname =~ /^(dst|src)$/)
	    {
		if (not _parse_rule_target($child, $filterEntry, $in))
		{
		    # omit generating rule if src or dst came back totally undefined, to avoid a rule which is an "any"
		    $LOGGER->debug("ZipTie::Adapters::Checkpoint::OPSEC::_process_rule : unable to resolve any source or dest for rule $name, property $cname; omitting filter entry generation");
		    return (undef, 0);
		}
	    }
	    elsif ($cname eq 'name')
	    {
	    	
	    	# The examples I've seen here start with hash signs we don't want to propagate
	    	# to a name override.
	    	if (defined $child->{simple} and $child->{simple} =~ /^\#*(.+)/)
	    	{
	    		# override name
	    		$filterEntry->{name} = $1;
	    	}
	    	
	    }
	    elsif ($cname eq 'service')
	    {
	        foreach my $ref (@{$child->{children}})
	        {
	            my $refname = _get_simple_by_name($ref->{children}, 'refname');
	            if ($refname)
	            {
            		$refname =~ s/^"//;
            		$refname =~ s/"$//;
            		$refname =~ s/^#_//;
            
            		if (defined $in->{portServiceGroups}->{"$refname"})
            		{
            			push (@{$filterEntry->{'destinationService'}}, { objectGroupReference => $refname, });
            		}
            		elsif (defined $in->{ipProtocolGroups}->{"$refname"})
            		{  # it's defined as a protocol
            			push (@{$filterEntry->{protocolObjectGroupReference}}, $refname);
            		}
            		
            		# Back to definition as a service -- check for source port mapping
            		if (defined $in->{portServiceGroups}->{"$refname" . "__src"})
            		{
            			push (@{$filterEntry->{'sourceService'}}, { objectGroupReference => $refname . "__src", });
            		}
	            }
	        }
	    }
	    elsif ($cname eq 'services')
	    {
	        foreach my $service (@{$child->{children}})
	        {
	            unless (defined $service->{name})
	            {
	                my $s = $service->{simple};
	                # Don't bother constraining if this is 'any', and omit any undefined
	                # object group references
	                unless ($s eq 'Any' or not defined($in->{portServiceGroups}->{"$s"}))
	                {
            	        push (@{$filterEntry->{'destinationService'}}, { objectGroupReference => $s, });

	            		# Check for source port mapping
	            		if (defined $in->{portServiceGroups}->{"$s" . "__src"})
	            		{
	            			push (@{$filterEntry->{'sourceService'}}, { objectGroupReference => $s . "__src", });
	            		}

	                }
	            }
	        }
	    }
	}
	if( ! $filterEntry->{ztSignature} )
	{
		$filterEntry->{ztSignature} = $name;
	}

    $action =~ s/^accept$/permit/;
    $action =~ s/^reject$/deny/;

	$filterEntry->{primaryAction} = $action;

    return ($filterEntry, $rPos);
}

sub _get_simple_by_name
{
    my ($children, $name) = @_;

    foreach my $child (@$children)
    {
        if ($child->{name} eq $name)
        {
            if (defined $child->{simple})
            {
                return $child->{simple};
            }

            return undef;
        }
    }

    return undef;
}

sub _parse_rule_target
{
    my ($target, $filterEntry, $in) = @_;
    my $zedTarget = ($target->{name} eq 'src') ? 'sourceIpAddr' : 'destinationIpAddr';

    foreach my $child (@{$target->{children}})
    {
        unless (defined $child->{name})
        {
            my $net = $child->{simple};
            if ($net eq 'Any')
            {
                push (@{$filterEntry->{$zedTarget}},  get_any());
            }
            else
            {
		# see if this shows up as a network group, omit mapping if if it does not.
		if (defined $in->{networkGroups}->{$net})
		{
		    my $objectGroup = { objectGroupReference => $net, };
		    push (@{$filterEntry->{$zedTarget}},  $objectGroup);
		}
		else
		{
		    $LOGGER->debug("ZipTie::Adapters::Checkpoint::OPSEC::_process_rule_target : cannot resolve $zedTarget spec (\"$net\") on $filterEntry->{name}, will try next (if any)");
		}
            }
        }
    }
    # Provide an indication as to what we mapped
    return ($filterEntry->{"$zedTarget"});
}

# If this is a host/gateway/gateway cluster object it creates an entry in $in for later use
sub _parse_checkpoint_objects {
	my ($in, $o, $name) = @_;

	my $type = $o->{type};
	if( $type eq 'host' or $type eq 'gateway' ) {
		if( $o->{cp_products_installed} && $o->{cp_products_installed} eq 'true' && $o->{ipaddr} )
		{
			push @{ $in->{cp_products_hosts} }, $o->{ipaddr};
		}

		# An object representing the local gateway will exist, this finds the IP address
		# assigned to an interface that matches this gateway object
		if( not $in->{my_gateway_metadata} and $o->{ipaddr} ) {
			my $local_addresses = _get_local_addresses($in);
			for my $local_address ( @{ $local_addresses } ) {
				if( $local_address eq $o->{ipaddr} )
				{
					my $interfaces;
					while( my ($interface_id, $interface) = each %{ $o->{interfaces} } )
					{
						$interfaces->{$interface->{officialname}} = {
							ipaddr => $interface->{ipaddr},
							mask   => $interface->{netmask},
							id     => $interface_id
						};
					}

					$in->{my_gateway_metadata} = {
						name       => $name,
						my_ip      => $o->{ipaddr},
						interfaces => $interfaces
					};
				}
			}
		}
	} elsif( $type eq 'gateway_cluster') {
		if( $o->{cluster_members} ) {
			my $cluster_map = $in->{clusters};
			while( my ($member_id, $member) = each %{ $o->{cluster_members} } )
			{
				$member->{refname} =~ /#_([^\"]*)/;
				push @{ $cluster_map->{$name}->{members} }, $1;
			}

			my $cluster_interfaces;
			while( my ($interface_id, $interface) = each %{ $o->{interfaces} } )
			{
				my $interface_name = $interface->{officialname};
				$cluster_interfaces->{ $interface_name } = {
					ipaddr => $interface->{ipaddr},
					mask   => $interface->{member_network}->{netmask},
					id     => $interface_id
				};
			}

			$cluster_map->{$name}->{interfaces} = $cluster_interfaces;
			$in->{clusters} = $cluster_map;
		}
	}

	return;
}

sub _get_my_cluster_interfaces {
	my ($in,$out) = @_;
	my $failovers;

	if ($in->{my_gateway_metadata} and $in->{clusters})
	{
		while( my ($name, $data) = each %{ $in->{clusters} } )
		{
			foreach my $member (@{$data->{members}})
			{
				if( $member eq $in->{my_gateway_metadata}->{'name'} )
				{
					# I belong to this cluster, build interface data. (keep going, I can belong to multiple clusters)
					while( my ($interface_name,$interface) = each(%{$data->{interfaces}}) )
					{
						$failovers->{$interface_name} =	{ 
							clusterxl => {
								failoverGroup => {
									groupID => $interface->{id},
									groupName => $name,
									primaryIPaddress => $interface->{ipaddr},
								},
							},
						};
					}
				}
			}
		}		
	}
	# otherwise we have no gateway and/or cluster data--no meaningful map of cluster i/f's available.
	
	return $failovers;
}

sub _parse_network_group {
	my ( $in, $o, $name ) = @_;
	my $network_group = { id => $name };

	my $type = $o->{type};
	if( $type eq 'machines_range' and $o->{ipaddr_first} and $o->{ipaddr_last} )
	{
		push @{$network_group->{range}}, {
			startAddress => $o->{ipaddr_first},
			endAddress   => $o->{ipaddr_last}
		};
	}
	elsif( $type eq 'host' and $o->{ipaddr} )
	{
		push @{$network_group->{host}}, $o->{ipaddr};
	}
	elsif( $type eq 'network' and $o->{ipaddr} and $o->{netmask} )
	{
		push @{$network_group->{subnet}}, {
			address => $o->{ipaddr},
			mask    => mask_to_bits( $o->{netmask} )
		};
	}
	elsif( $type eq 'gateway' and ( $o->{ipaddr} or $o->{interfaces} ) )
	{
		if( $o->{interfaces} ) {
			for my $ifName ( keys %{ $o->{interfaces} } )
			{
				my $interface = $o->{interfaces}->{$ifName};

				# watch for unconfigured interfaces
				if( $interface->{ipaddr} )
				{
					push @{ $network_group->{host} }, $interface->{ipaddr};
				}

				# Keep the ifIndex for later printing within parse_interface
				if( defined $interface->{ifindex} ) {
					my $official_name = $interface->{officialname};
					$in->{gateways}->{$name}->{interfaces}->{$official_name}->{ifindex} = $interface->{ifindex};
				}
			}
		} elsif( $o->{ipaddr} ) {
			push @{ $network_group->{host} }, $o->{ipaddr};
		}

		if( not $network_group->{host} ) {
			$LOGGER->debug("Gateway object '$name' found with no interface address");
			return;
		}
	}
	elsif( $type eq 'group' and $o->{member_class} and $o->{member_class} eq 'network_object' )
	{
		for my $child ( @{ $o->{children} } ) {
			push @{ $network_group->{objectGroupReference} }, $child;
		}				
	}
	elsif( $type eq 'group_with_exclusion' )
	{
		$network_group->{base} = $o->{base}->{Name};
		$network_group->{exclusion} = $o->{exception}->{Name};
		
	} else {
		if( $o->{valid_ipaddr6} or $o->{ipaddr6} or $o->{ipaddr_first6} ) {
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "IPv6 not supported for '$name'");
		} else {
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Invalid network object: '$name'");
		}

		$LOGGER->verbose("$name:\n" . Dumper($o));
		return;
	}

	return $network_group;
}

# Process group elements strictly defined by the IP protocols being referenced by them e.g. ICMP, OSPF, IGMP, etc
sub _parse_protocol_group {
	my ( $in, $o, $name ) = @_;
	my $protocol_group;

	# Looking for base protocols here.  That means that it is defined by its IP protocol,
	# with no other significant discriminator, such as a payload string to search for.
	if( $o->{protocol} && not $o->{exp} ) {
		my $protocol = $o->{protocol};

		# Is this an actual IP protocol?		
		if( $protocol > 0 ) {
			my $pName = get_protocol_name($protocol);
			
			if( $pName ) {
				$protocol_group = {
					id => $name,
					protocol => $pName
				};

				# Record this for rule reference posterity
				$in->{ipProtocolGroups}->{$name} = $protocol_group;
			}
		}
	}

	return $protocol_group;
}

sub _parse_service_group {
	my ( $in, $o, $name ) = @_;
	my $service_group;
	my $src_service_group;

	my $protocol = lc $o->{type};
	if ( $protocol eq 'group' ) {
		my @object_references;
		for my $child ( @{ $o->{children} } ) {

			# see if this was already generated as an elemental protocol group or port service.  If so, then
			# we can leverage it here, otherwise, we omit it to avoid an undefined reference
			if( defined $in->{ipProtocolGroups}->{$child} or defined $in->{portServiceGroups}->{$child} ) {
				push @object_references, $child;
			}

			# if we've already picked up some protocol refs, then we're either missing this one or its mixed somehow
			elsif( @object_references ) {
				$LOGGER->debug("Unable to map object '$child' in the process of assembling referencing service group '$name'");
			}
		}

		if( @object_references ) {
			$service_group = {
				id => $name,
				objectGroupReference => \@object_references
			};

			$in->{ipProtocolGroups}->{$name} = $service_group;
		}
	} elsif( $protocol eq 'tcp' or $protocol eq 'udp' ) {
		$service_group = { id => $name };
		my $port_expression = {};

		my $port = $o->{port};
		if( $port =~ /^([<>])?(\d+)$/ ) {
			my $operator = $1 ? $1 eq '>' ? 'gt' : 'lt' : 'eq';
			$port_expression = {
				port     => $2,
				operator => $operator,
				protocol => $protocol,
			};
			$service_group->{portExpression} = $port_expression;
		} elsif ( $port =~ /^(\d+)\s*-\s*(\d+)$/ ) {
			$port_expression = {
				portStart => $1,
				portEnd   => $2,
				protocol  => $protocol
			};
			$service_group->{portRange} = $port_expression;
		}
		$in->{portServiceGroups}->{$name} = $port_expression;

		my $srcPort = $o->{src_port};
		if( $srcPort )
		{
			my $src_name = $name . '__src';
			$src_service_group = { id => $src_name };
			my $src_port_expression = {};

			if ( $srcPort =~ /^([<>])?(\d+)$/ ) {
				my $operator = $1 ? $1 eq '>' ? 'gt' : 'lt' : 'eq';
				$src_port_expression = {
					port     => $2,
					operator => $operator,
					protocol => $protocol
				};
				$src_service_group->{portExpression} = $src_port_expression;
			} elsif ( $srcPort =~ /^(\d+)\s*-\s*(\d+)$/ ) {
				$src_port_expression = {
					portStart => $1,
					portEnd   => $2,
					protocol  => $protocol
				};
				$src_service_group->{portRange} = $src_port_expression;
			}

			# Make sure we know we had a source mapping here.		
			$in->{portServiceGroups}->{$src_name} = $src_port_expression;
		}
	}
	return ($service_group, $src_service_group);
}

sub _rule_to_hash
{
   my $object_text = shift;

   my $parser = ZipTie::Adapters::ConfigObjects->new();

   return $parser->run($object_text);
}

sub object_to_hash
{
	# turns a check point configuration object into a standard
	# perl hashtable
	my $objectText = shift;
	my $newObject  = {};

        # replace "( : () )" with "( )"
	$objectText =~ s/\(\s*:\s*\(\s*\)\s*\)/( )/;
	# fix some odd structure
	$objectText =~ s/(:[a-zA-Z0-9_-]+\s*\()([a-zA-Z0-9_-]+)\s*$/$1/mg;

	my @stack;
	push(@stack, $newObject);

	my $key;
	my $word;
	my $uptonow = "";
	my $stray_paren = 0;
	my $quotedContext = 0;

	foreach my $char ( split( //, $objectText ) )
	{
	    $uptonow .= $char;
		$word =~ s/\s*$// if ( $char =~ /[:()]/ );    # remove only trailing whitespace

		if ($char eq '"')
		{
			# Either turning on or turning off ignoring special chars.  N.B.!  I 
			# donc't know how quotes are escaped...."" ?  I couldn't find any examples.
			# It's not backslash, I can tell you that.
			$quotedContext = not $quotedContext;
		}
		elsif ( $char eq ':' and not $quotedContext)
		{
			
			if ($word)
			{
				if ( $word =~ /^\s/ && !$key )
				{
					$word =~ s/^\s+//;
					$word =~ s/\s+$//;
					my $current = peek(@stack);
					push( @{ $current->{'children'} }, $word );
				}
				else
				{
					my $current = peek(@stack);
					$current->{$word} = {};
					push( @stack, $current->{$word} );
				}
			}
			elsif ($uptonow =~ /:\s*\(\s*:$/)
			{
			    my $current = peek(@stack);
			    $current->{extra_paren} = 1;
			}
		}
		elsif ( $char eq '(' and not $quotedContext)
		{
		    $key = $word;
		    my $current = peek(@stack);
			if ( length($key) > 0 )
			{
				$current->{$key} = {};
				push( @stack, $current->{$key} );
			}
		}
		elsif ( $char eq ')' and not $quotedContext)
		{
			if ( $word =~ /^\s/ && !$key )
			{
				$word =~ s/^\s+//;
				$word =~ s/\s+$//;
				my $current = pop(@stack);
				push( @{ $current->{'children'} }, $word );
			}
			elsif (!$key && $uptonow =~ /\(\S+\)$/)
			{
			    my $current = peek(@stack);
			    $current->{$word} = {};
			}
			else
			{
			    my $c = peek(@stack);
			    if ( exists $c->{extra_paren} )
			    {
				delete $c->{extra_paren};
			    }
			    else
			    {
				
				pop(@stack);
				if ( defined $word && length($key) > 0 )
				{
					my $current = peek(@stack);
					$current->{$key} = $word;
				}
			    }

			}
		}
		else
		{
			$word .= $char;
		}

		# Clear key and word context, but only if we're not slogging through
		# a quoted string
		
		if (not $quotedContext)
		{
			$word = '' if ( $char =~ /[:()]/ );
			$key  = '' if ( $char =~ /[:)]/ );			
		}
	}

	# Finally, if we closed with a child initialized but didn't push it because input ended, do it
	# now.
	if ($word && $word =~ /\S/)
	{
		foreach my $newObjKey (keys(%{$newObject}))
		{
			# No child left behind!
			$word =~ s/^\s+//;
			$word =~ s/\s+$//;
			chomp($word);
			push( @{ $newObject->{$newObjKey}->{'children'} }, $word );			
			last;
		}
		
	}
	return $newObject;
}

sub peek
{
	my @array = @_;
	return $array[$#array];
}

sub get_any
{
	my @anyObj;
	push (@anyObj, { network => get_any_address_encoding()} );
	return @anyObj; 
}

sub parse_interfaces {
	my ($in, $out) = @_;
	my @interfaces;

	my $cluster_interfaces = _get_my_cluster_interfaces($in,$out);
	my $ifIndex;

	while( $in->{interfaces} =~ /^(\S+)\s+Link(.+?)(?=^\s*$)/msg ) {
		my ($name, $blob) = ($1, $2);

		$LOGGER->debug("Parsing interface '$name'");
		$LOGGER->verbose("Interface blob:\n'$blob'");

		my $interfaceType = get_interface_type($1);
		if( $interfaceType eq 'other' and $blob =~ /^\s*encap:Ethernet/ ) {
			$interfaceType = 'ethernet';
		}

		my $interface = {
			name          => $name,
			adminStatus   => 'down',
			interfaceType => $interfaceType,
		};

		if( $blob =~ /^\s*UP/m ) {
			$interface->{adminStatus} = 'up';
		}

		$ifIndex = _getIfIndex($in,$interface);
		if (defined $ifIndex) {
			$interface->{ifIndex} = $ifIndex;	
		}

		$interface->{physical} = $interface->{interfaceType} eq 'softwareLoopback' ? 'false' : 'true';
		( $interface->{mtu} ) = $blob =~ /MTU:(\d+)/;

		if ( $blob =~ /HWaddr\s+([a-z\d:-]+)/i )
		{
			my $mac = $1;
			$mac =~ s/[:-]//g;

			unless ($mac =~ /^0+$/)
			{
				$interface->{interfaceEthernet}->{macAddress} = $mac;
			}
		}

		while( $blob =~ /^\s+inet\s+addr:(\S+)\s+(.+)/mg ) {
			my $address = $1;
			my $line = $2;

			my $ip_configuration = { ipAddress => $address };

			if( $line =~ /Mask:(\S+)/i ) {
				$ip_configuration->{mask} = mask_to_bits($1);
			}
			if( $line =~ /Bcast:(\S+)/i ) {
				$ip_configuration->{broadcast} = $1;
			}

			push @{ $interface->{interfaceIp}->{ipConfiguration} }, $ip_configuration;

			if( !$interface->{failover} ) {
				my $cidr = new Net::CIDR::Lite;

				# TODO refactor this statement
				$cidr->add( ($ip_configuration->{ipAddress} . ($ip_configuration->{mask}?"/".$ip_configuration->{mask}:"")) );

				while( my ($cname,$failover) = each(%{$cluster_interfaces}) )
				{
					if( $cidr->find($failover->{clusterxl}->{failoverGroup}->{primaryIPaddress}) )
					{
						$interface->{failover} = $failover;
					}
				}
			}
		}

		$LOGGER->verbose("Parsed inteface: " . Dumper($interface));
		push @interfaces, $interface;
	}

	if( @interfaces ) {
		$out->print_element( 'interfaces', { interface => \@interfaces } );
	}

	return;
}

# Parses and returns all the ipv4 addresses assigned to interfaces
sub _get_local_addresses {
	my $in = shift;

	# If already parsed then just return
	return $in->{local_addresses} if $in->{local_addresses};

	my @local_addresses = ();
	while( $in->{interfaces} =~ /^\s+inet\s+addr:(\S+)/mg ) {
		push @local_addresses, $1;
	}

	$LOGGER->verbose("Parsed local addresses:\n" . Dumper(\@local_addresses));

	$in->{local_addresses} = \@local_addresses;

	return \@local_addresses;
}

# this function sets interface's index
sub _getIfIndex {
	my ($in, $interface) = @_;
	
	unless (defined $in->{hostname}){
		return undef;
	}
		
	my ($hostname) = $in->{hostname} =~ /hostname\s+(\S+)/g;
	
	my $ifIndex;
	
	if (not defined $in->{gateways}->{$hostname}->{interfaces}->{$interface->{name}}){
		return undef;
	} else {
		$ifIndex = $in->{gateways}->{$hostname}->{interfaces}->{$interface->{name}}->{ifindex};	
	}
	
	return $ifIndex;
}

sub _count_rule_set
{
	my $ruleSet = shift;
	my $ruleCount = 0;
	
	for (keys(%{$ruleSet}))
	{
		$ruleCount += scalar ( @{$ruleSet->{$_}});
	}
	
	return $ruleCount;
}

# This function parse objects.C file and calls getNetworkObject (to store network objects elements)
# or getFirewallProperty function (to store firewall_properties)
sub _parse_objectsFile{
	my ( $in, $out ) = @_;
	
	unless (exists $in->{objectsFile}){
		return;
	}

	open( OBJECTS, $in->{objectsFile} );
	my $parentWhitespace;           # leading whitepsace on the netobj definition
	my $currentObjectWhitespace;    # leading whitespace on each object definition
	my $currentObject;
	my $type;
	my $fw_prop; #firewall property flag
	
	while ( my $line = <OBJECTS> )
	{
    	# Handle R70 comments convention
 		if( $line =~ /^\s*#.*$/ )
		{
			next;
		}
		
		if ( $line =~ /(^\s+):(netobj|props) \(/ ){
			$type             = $2;
			if ( ($type eq 'netobj') or ( ($type eq 'props') and ($line =~ /(^\s+):(props)\s+\(firewall_properties/) )){
				$parentWhitespace = $1;
				if ($line =~ /firewall_properties/) {
					$fw_prop = 1;
				}
			}
		}
		elsif ($parentWhitespace)
		{
			if ( $line =~ /^$parentWhitespace\)/ )
			{
				$parentWhitespace = undef;
			}
			elsif ( ( $line =~ /(^\s+):\s*\(/ ) && ( !$currentObjectWhitespace ) )
			{
				$currentObjectWhitespace = $1;
				$currentObject .= $line;
			}
			elsif ( $currentObjectWhitespace && ( $line =~ /^$currentObjectWhitespace\)/ ) )
			{
				if ( $type eq 'netobj' ){
					_getNetworkObjects( $in, $currentObject, $out );
					$currentObject           = undef;
					$currentObjectWhitespace = undef;
				}
				elsif ($type eq 'props' ){
					# continue to build currentObject's content
					$currentObject .= $line;	
				}
			}
			else{
				$currentObject .= $line;
			}
		}
	}
	
	if ( ($type eq 'props') and (defined $fw_prop) ){
		_getFirewallProperties($in, $currentObject, $out)
	} 
	
	close OBJECTS
}

#this function stores network Objects
sub _getNetworkObjects{
	my ( $in, $object, $out ) = @_;

	my $networkGroup = object_to_hash($object);
	my $name         = ( keys(%$networkGroup) )[0];
    my $o            = $networkGroup->{$name};

    return unless (ref $o);
	
	my $type         = $o->{'type'};
	if ( $type && $type =~ /^(machines_range|host|network|group)$/ ){	
		#store networkGroup
		push (@{$in->{networkObjects}}, $networkGroup);
	}
}

# this function stores firewall properties object
sub _getFirewallProperties {
	my ( $in, $object, $out ) = @_;

	my $firewallProperty = object_to_hash($object);
	       
	#store firewall property    
   	$in->{firewall_properties} = $firewallProperty;
}

# parse network_objects to check NAT's existence.
# Looks for firewalling_obj elements and stores NAT attributes of the network objects
sub checkNAT {
	my ($in, $out) = @_;
	
	my $hostname = "";
	if (defined $in->{hostname}){
		if ($in->{hostname} =~ /hostname\s+(\S+)/){
			$hostname = $1;
		}
	}
	
	$in->{isNAT} = 0;
	
	# clean previous $in->{networkObjects}
	if (exists $in->{networkObjects}){
		delete $in->{networkObjects};
	}
	
	# clean previous $in->{firewall_properties}
	if (exists $in->{firewall_properties}) {
		delete $in->{firewall_properties};
	}
	
	_parse_objectsFile($in,$out);

	foreach my $netObj (@{$in->{networkObjects}}){
		
		#my $networkObject = object_to_hash($netObj);
		my $name         = ( keys(%$netObj) )[0];
	    
	    #return unless (ref $object_name);
	    
	    if( ($name ne 'CP_default_Office_Mode_addresses_pool') and exists($netObj->{$name}->{the_firewalling_obj}) ){	
	    	
    		# store NAT settings for future references
			# NAT rules are installed on the Gateway specified in the_firewalling_obj
			if( defined($netObj->{$name}->{the_firewalling_obj}) and (ref($netObj->{$name}->{the_firewalling_obj}) eq 'HASH') ) {
				if( (exists $netObj->{$name}->{the_firewalling_obj}->{refname}) and ($netObj->{$name}->{the_firewalling_obj}->{refname} =~ /^#_(\S+)/) ){ 
					$in->{NAT}->{$name}->{the_firewalling_obj} = $1;	
				} 	
			}
			
			# store NAT translation method (i.e static or hide (dynamic) )
			$in->{NAT}->{$name}->{netobj_adtr_method} = defined($netObj->{$name}->{netobj_adtr_method}) ? $netObj->{$name}->{netobj_adtr_method}: undef;
			# valid address name
			$in->{NAT}->{$name}->{valid_addr_name} = defined($netObj->{$name}->{valid_addr_name}) ? $netObj->{$name}->{valid_addr_name}: undef;
			# valid NAT IP address
			$in->{NAT}->{$name}->{valid_ipaddr} = defined($netObj->{$name}->{valid_ipaddr}) ? $netObj->{$name}->{valid_ipaddr}: undef;
				
			# found a network_objects that has a reference to the analyzed gateway
			if( ($in->{isNAT} == 0 and defined($in->{NAT}->{$name}->{the_firewalling_obj}) ) 
					and (	($in->{NAT}->{$name}->{the_firewalling_obj} =~ /All/) or 
							($in->{NAT}->{$name}->{the_firewalling_obj} =~ /Gateways/) or
							($in->{NAT}->{$name}->{the_firewalling_obj} =~ /$hostname/)   
							 ) ){
						$in->{NAT}->{installed_on_gateway} = defined ($in->{my_gateway_metadata}) ? $in->{my_gateway_metadata} : undef;						
						$in->{isNAT} = 1;
					}
	    		
	    } # end ne 'CP_default_Office_Mode_addresses_pool'	
	} #end foreach in->{networkObjects}

	return;
}

# Prints a network group with valid ordering of elements
sub _print_network_group{
	my ($out, $network_group) = @_;
	my @elements = qw(objectGroupReference host range subnet);

	$out->open_element('networkGroup');
	$out->print_element('id', $network_group->{id});
	for my $element ( @elements ) {
		if(defined $network_group->{$element}){
			$out->print_element($element, $network_group->{$element});
		}
	}
	$out->close_element('networkGroup');

	return;
}

sub _get_network_set {
        my ($name, $network_groups) = @_;
        my $set;

        if( exists $network_groups->{$name} ) {
                my %network_group = %{ $network_groups->{$name} };
                $set = Net::CIDR::Set->new();
                for my $key ( keys %network_group ) {
                        my $value = $network_group{$key};
                        if( $key eq 'objectGroupReference' ) {
                                for my $ref ( @{ $value } ) {
                                        my $nested_set = _get_network_set($ref, $network_groups);
                                        if( $nested_set ) {
                                                $set->merge($nested_set);
                                        }
                                }
                        } elsif( $key eq 'host' ) {
                                for my $host ( @{ $value } ) {
                                        $set->add($host);
                                }
                        } elsif( $key eq 'subnet' ) {
                                for my $subnet ( @{ $value } ) {
                                        my $ip = $subnet->{address};
                                        my $mask = $subnet->{mask};
                                        $set->add("$ip/$mask");
                                }
                        } elsif( $key eq 'range' ) {
                                for my $range ( @{ $value } ) {
                                        my $start = $range->{startAddress};
                                        my $end = $range->{endAddress};
                                        $set->add("$start-$end");
                                }
                        }
                }
        }

        return $set;
}


1;
