package ZipTie::Adapters::CheckPoint::SecurePlatform::Parsers;

use 5.010_000;
use strict;
use warnings;
use ZipTie::Adapters::Utils qw(seconds_since_epoch mask_to_bits get_port_number trim);
use ZipTie::Adapters::Unix::ParsingUtils;
use ZipTie::Adapters::CheckPoint::OPSEC::Parsers qw(checkNAT);
use ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser;
use ZipTie::Logger;
use MIME::Base64 'encode_base64';

use Exporter 'import';
our @EXPORT_OK =
  qw(create_config parse_local_accounts parse_chassis parse_snmp parse_system parse_static_routes);

my $LOGGER = ZipTie::Logger::get_logger();
my $make = "Check Point";

sub parse_system {
	my ($in, $out) = @_;

	# Parse host name or fail backup
	if( $in->{hostname} =~ /hostname\s+(\S+)/ ){
		$out->print_element('core:systemName', $1);
	} else {
		$LOGGER->fatal('Hostname is required for a successful backup!');
	}

	# Parse OS info
	$out->open_element('core:osInfo');
	$out->print_element('core:make', $make);

	my $version = 'N/A';
	my $os_type = 'N/A';
	if( $in->{version} =~ /Product\s+version\s+Check\s+Point\s+(\S+)\s(\S+)/ ) {
		($os_type, $version)  = ($1, $2);
	}
	$out->print_element('core:version', $version); # required element
	$out->print_element('core:osType',  $os_type);

	# Print packet processing order if there is NAT
	checkNAT($in, $out);
	if ($in->{isNAT}) {
		my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
		my $packetProcessingOrdering = $addressTranslationParser->getPacketProcessingOrdering();
		$out->print_string($packetProcessingOrdering);
	} 
	$out->close_element('core:osInfo');

	# Parse BIOS version
	if( $in->{dmidecode} =~ /(BIOS Information.+?)(?=^\S)/ms ) {
		if( $1 =~ /Version:\s*(\S+)/ ) {
			$out->print_element('core:biosVersion', $1);
		}
	}

	$out->print_element( 'core:deviceType', 'Firewall' );

	# Parse contact email
	if( $in->{snmp} =~ /^syscontact\s*(\b.+\b)/m ) {
		$out->print_element('core:contact', $1);
	}

	ZipTie::Adapters::Unix::ParsingUtils::parse_last_reboot( $in->{uptime}, $out ) if defined ($in->{uptime});
	return;
}

sub parse_chassis {
	my ($in, $out) = @_;

	my $asset = {
		'core:assetType' => 'Chassis',
		'core:factoryinfo' => {
			'core:make' => $make,
			'core:modelNumber' => 'Security Gateway'
		}
	};

	# Append version to model if possible
	if( $in->{version} and $in->{version} =~ /Check\s+Point\s+\S+\s(\S+)/ ) {
		$asset->{'core:factoryinfo'}->{'core:modelNumber'} .= " $1";
	}

	# Parse serial number if available
	if( $in->{dmidecode} =~ /System\s+Information.+?Serial\s+Number:[^\S\n]+(\S.+?)\s*$/ism ) {
		$asset->{'core:factoryinfo'}->{'core:serialNumber'} = $1;	
	}

	my $chassis = { 'core:asset' => $asset };
	$out->print_element('chassis', $chassis);

	return;
}

sub _add_folder
{
    my ($parent, $name) = @_;

	my $folder = { 'core:name' => $name, };
	push( @{ $parent->{'core:folder'} }, $folder );
	return $folder;
}

sub create_config
{
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository = { 'core:name' => '/', };
	my $fw_folder = $repository;
	foreach my $dirName ( split /\//, $in->{fwdir} )
	{
		if ( length($dirName) )
		{
		    $fw_folder = _add_folder($fw_folder, $dirName )
		}
	}


	# for slurping the entire file at once
	my $lineEnding = $/;
	undef $/;   
	my $wholeFile;

	open (FILE, $in->{rulesFile}) or die("Unable to open rulesFile (" . $in->{rulesFile} . "): $!");
	$wholeFile = <FILE>;
	close (FILE);
	my $rules = {
		'core:name'       => 'rules.C',
		'core:textBlob'   => encode_base64($wholeFile),
		'core:mediaType'  => 'text/plain',
		'core:context'    => 'active',
		'core:promotable' => 'true',
	};
	
	open (FILE, $in->{objectsFile}) or die("Unable to open objectsFile (" . $in->{objectsFile} . "): $!");
	$wholeFile = <FILE>;
	close (FILE);

	my $objects = {
		'core:name'       => 'objects.C',
		'core:textBlob'   => encode_base64($wholeFile),
		'core:mediaType'  => 'text/plain',
		'core:context'    => 'active',
		'core:promotable' => 'true',
	};

	undef $wholeFile;

	$/ = $lineEnding;

	# push the configuration into the repository
    my $db_folder = _add_folder($fw_folder, 'database');
	push( @{ $db_folder->{'core:config'} }, $rules );
	push( @{ $db_folder->{'core:config'} }, $objects );

	# print the repository
	$out->print_element( 'core:configRepository', $repository );
}

sub parse_local_accounts
{
	my ( $in, $out ) = @_;
	$out->open_element("localAccounts");
	while ($in->{users} =~ /^(\S+)/mg)
	{
		my $name = $1;
		if ($name !~ /showusers|#|@/)
		{
			my $account = { accountName=>$name, accessGroup=>'admin', };
			$out->print_element('localAccount', $account);
		}
	}
	$out->close_element("localAccounts");
	
}

sub parse_snmp
{
	my ( $in, $out ) = @_;
	$out->open_element("snmp");
	my @communities = ();
	my @traps       = ();

	# parse communities and trap hosts
	while ( $in->{snmp} =~ /^(.+)$/mig )
	{
		my $commline = $1;
		if ( $commline =~ /^(r[ow])community\S*\s+(\S+)\s+\S+\s*(\S*)$/i )
		{
			push( @communities, { accessType => uc($1), communityString => $2, mibView => $4 } );
		}
		elsif ( $commline =~ /^trapsess.+\-c\s+(\S+)\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).+$/i )
		{
			push( @traps, { ipAddress => $2, communityString => $1 } );
		}
	}

	# parse other snmp data and store the configuration
	foreach (@communities)
	{
		$out->print_element( "community", $_ );
	}

	my $someSysPrinted = 0;
	if ( $in->{snmp} =~ /^syscontact\s+"?(\b.+\b)/mi )
	{
		$out->print_element( "sysContact", $1 );
		$someSysPrinted = 1;
	}

	if ( $in->{snmp} =~ /^syslocation\s+"?(\b.+\b)/mi )
	{
		$out->print_element( "sysLocation", $1 );
		$someSysPrinted = 1;
	}

	$in->{hostname} =~ s/hostname\s*//g;
	if ( $in->{hostname} =~ /(\S+)/ )
	{
		$out->print_element( "sysName", $1 );
		$someSysPrinted = 1;
	}

	if ($someSysPrinted)
	{
		foreach (@traps)
		{
			$out->print_element( "trapHosts", $_ );
		}
	}

	$out->close_element("snmp");
}

sub parse_static_routes
{
	my ( $in, $out ) = @_;
	my $staticRoutes;

	# grab all data about static route
	while ( $in->{routes} =~ /([\d\.]+)\s+([\d\.]+)\s+([\d\.]+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)/g )
	{
		if (   $1 ne '127.0.0.0'
			&& $1 ne '127.0.0.1' 
			&& $2 ne '0.0.0.0')
		{
			my $route = {
				destinationAddress => $1,
				destinationMask    => mask_to_bits($3),
			};
			$route->{gatewayAddress} = $2;
			$route->{interface}      = $8;
			if ( ( $route->{destinationAddress} eq '0.0.0.0' ) && ( $route->{destinationMask} eq '0' ) )
			{
				$route->{defaultGateway} = 'true';
			}
			else
			{
				$route->{defaultGateway} = 'false';
			}
			$route->{routeMetric} = $5;
			push( @{ $staticRoutes->{staticRoute} }, $route );
		}
	}
	$out->print_element( 'staticRoutes', $staticRoutes ) if ( defined $staticRoutes );
}

1;
