package ZipTie::RESTAPIv2;

use strict;
use warnings;

use LWP::UserAgent;
use ZipTie::ConnectionPath;

use ZipTie::Recording;
use ZipTie::Recording::Interaction;
my $RECORDING = ZipTie::Recording::get_recording();

use ZipTie::Logger;
my $LOGGER = ZipTie::Logger::get_logger();
my $QUIET = 1;

# Creates a new REST API object from a connectionPath Object
sub new {
	my ($class, $connection_path, $api) = @_;

	$LOGGER->debug("Creating RESTAPI Object");

	my $connection_protocol = $connection_path->get_protocol_by_name('HTTPS');
	my $protocol = lc $connection_protocol->get_name();

	# Could be a proxied backup, check if there is a management server
	my $ip = $connection_path->get_management_server_ip_address() // $connection_path->get_ip_address();
	my $port = $connection_protocol->get_port();
	my $url = "$protocol://$ip:$port";

	my $client = LWP::UserAgent->new();
	$client->protocols_allowed( [ $protocol ] );
	my $timeout = 300;
	$client->timeout($timeout);

        $LOGGER->debug("URL: $url");
        $LOGGER->debug("API Path: $api");
	$LOGGER->debug("Timeout: $timeout");

	my $this = {
		client => $client,
		url => $url,
		api => $api
	};

	bless $this, $class;
	return $this;
}

# Wrapper for a LWP::UserAgent->get call
# Logs the url to get, the response content and time taken
# Also handles creation of a RESTAPI recording
sub _get {
	my ($this, $name, $url, $content, $quiet) = @_;

	_start_recording( $url, $name );

	if( not $quiet ) {
		$LOGGER->debug("Requesting Web Service: $url");
		$LOGGER->debug("Request body: $content");
	}

	my $time = time;
	my $headers = $this->{headers} // [];

	# URL, HTTP Headers, POST body passed to user agent
	my $response = $this->{client}->post( $url, @{ $headers }, 'Content' => $content );
	$response = $response->content();

	$time = time - $time;

	if( not $quiet ) {
		$LOGGER->debug("Response Received: $response");
		$LOGGER->debug("Response Time: $time seconds");
	}

	_end_recording( $response );

	return $response;
}

# Calls to get an authentication token
sub _obtain_token {
	my ($this, $endpoint, $content, $token_parser) = @_;
	$LOGGER->debug('Sending Authentication Request');

	my $url = $this->{url} . $this->{api} . $endpoint;
	my $response = $this->_get('obtain connection token', $url, $content, $QUIET);
	my $token = &{$token_parser}( $response );

	$LOGGER->debug('Authentication Successful');
	return $token;
}

# Authentication tokens can be passed as a header
sub obtain_token_as_header {
	my ($this, $endpoint, $content, $token_parser, $token_header) = @_;
	my $token = $this->_obtain_token($endpoint, $content, $token_parser);
	$this->{headers} = [ @{ $this->{headers} }, $token_header => $token ];
	return $token;
}

# Authentication tokens can be passed as a url parameter
sub obtain_token_as_parameter {
	my ($this, $endpoint, $content, $token_parser) = @_;
	my $token = $this->_obtain_token($endpoint, $content, $token_parser);
	$this->{token} = $token;
	return $token;
}

# The content of the HTTP response is returned for a given API endpoint
# based off <PROTOCOL>://<IP>:<PORT><API><COMMAND>
sub get {
	my ($this, $name, $query, $content) = @_;

	my $url = $this->{url} . $this->{api} . $query;

	# If authentication token is passed as url parameter prepend
	if( $this->{token} ) {
		$url .= $this->{token};
	}

	my $response = $this->_get($name, $url, $content);

	return $response;
}

# The content of the HTTP response is returned for a given API endpoint
# based off <PROTOCOL>://<IP>:<PORT><ENDPOINT>
sub get_custom {
	my ($this, $name, $endpoint, $content) = @_;

	my $url = $this->{url} . $endpoint;

	# If authentication token is passed as url parameter prepend
	if( $this->{token} ) {
		$url .= $this->{token};
	}

	my $response = $this->_get($name, $url, $content);

	return $response;
}

# This function sets up and invokes the start_current_interaction function of recording.
sub _start_recording {
	my ($command, $name) = @_;

	if ( $ENV{ENABLE_RECORDING} ) {

		# If the command contains xml tags, replace the angle braces with percent
		# representation so that the recording matches the url that gets sent
		$command =~ s/\</\%3C/msgx;
		$command =~ s/\>/\%3E/msgx;
		# TODO: Use URI::Encode to handle all symbols not just < and >

		my $interaction = ZipTie::Recording::Interaction->new( richCommand => $command );
		if ( defined $name ) {
			$interaction->name($name);
		}
		$RECORDING->start_current_interaction($interaction);
	}

	return;
}

# This function sets up and invokes the finish_current_interaction function of recording.
sub _end_recording {
	my $response = shift;

	if ( $ENV{ENABLE_RECORDING} ) {
		my $interaction = $RECORDING->get_current_interaction();
		$interaction->cliResponse($response);
		$RECORDING->finish_current_interaction();
	}

	return;
}

1;
